;=========================== DOSEXT FILE functions ===========================

;
; Create file
; In:
;   ESI -> ASCIIZ filename
; Out:
;   CF=1 - Error creating file
;   CF=0 - File created succesfully
;     REGS._BX - file handle
;
Createfile proc near
        Push eax ebx edi esi
        Add esi,codeaddx
        Mov eax,esi
        Mov edi,Offset Regs
        And esi,0Fh
        Shr eax,4
        Mov [edi]._dx,si
        Mov [edi]._ds,ax
        Mov [edi]._ah,3Ch
        Mov [edi]._cx,0020h
        Mov bl,21h
        Call RealInt
        Mov ax,[edi]._ax
        Mov [edi]._bx,ax
        BT dword ptr [edi]._flag,0
        Pop esi edi ebx eax
        Ret
Createfile endp

;
; Open file
; In:
;   ESI -> ASCIIZ filename
; Out:
;   CF=1 - Error opening file
;   CF=0 - File opened succesfully
;     REGS._BX - file handle
;
Openfile proc near
        Push eax ebx edi esi
        Add esi,codeaddx
        Mov edi,Offset Regs
        Mov eax,esi
        And esi,0Fh
        Shr eax,4
        Mov [edi]._dx,si
        Mov [edi]._ds,ax
        Mov [edi]._ax,3D02h
        Mov bl,21h
        Call RealInt
        Mov ax,[edi]._ax
        Mov [edi]._bx,ax
        BT dword ptr [edi]._flag,0
        Pop esi edi ebx eax
        Ret
Openfile endp

;
; Close a file
; In:
;   REGS._BX - file handle
; Out:
;   None
;
Closefile proc near
        Push eax ebx
        Mov ah,3Eh
        Mov bx,regs._bx
        Int 21h
        Pop ebx eax
        Ret
Closefile endp

;
; Delete a file
; In:
;   ESI -> ASCIIZ filename
; Out:
;   CF=1 - Error deleting file
;   CF=0 - File deleted succesfully
;
Deletefile proc near
        Push eax ebx edi esi
        Add esi,codeaddx
        Mov edi,Offset Regs
        Mov eax,esi
        And esi,0Fh
        Shr eax,4
        Mov [edi]._dx,si
        Mov [edi]._ds,ax
        Mov [edi]._ah,41h
        Mov [edi]._cx,0020h
        Mov bl,21h
        Call RealInt
        BT dword ptr [edi]._flag,0
        Pop esi edi ebx eax
        Ret
Deletefile endp

;
; Seek position in file
; In:
;   REGS._BX - file handle
;   EAX - signed offset to move to
;   BL - from: 0-beginning of file, 1-current location, 2-end of file
; Out:
;   CF=1  - Error seeking in file
;     EAX - ?
;   CF=0  - Seek fine
;     EAX - new offset from beginning of file
;
Seekfile proc near
    Push ebx ecx edx

    Mov dx, ax
    Shr eax, 16
    Mov cx, ax

    Mov ah, 42h
    Mov al, bl
    Mov bx, regs._bx
    Int 21h

    Pushf
    Mov cx, ax
    Mov ax, dx
    Shl eax, 16
    Mov ax, cx
    Popf

    Pop edx ecx ebx
    Ret
Seekfile endp

;
; Get size of file
; In:
;   REGS._BX - file handle
; Out:
;   CF=1  - Error checking file
;     EAX - ?
;   CF=0  - chek fine
;     EAX - size of file
;
Filesize proc near
        Push ebx ecx edx
        Mov ax,4201h
        Xor ecx,ecx
        Xor edx,edx
        Mov bx,regs._bx
        Int 21h
        Push edx eax
        Mov ax,4202h
        Xor ecx,ecx
        Xor edx,edx
        Int 21h
        Xchg ax,dx
        Shl eax,16
        Mov ax,dx
        Pop edx ecx
        Push eax
        Mov ax,4200h
        Int 21h
        Pop eax edx ecx ebx
        Ret
Filesize endp

;
; Read from file
; In:
;   REGS._BX - file handle
;   EDI -> buffer to read to
;   ECX - number of bytes to read
; Out:
;   CF=1 - Error reading file
;     EAX - ?
;   CF=0 - Read went fine
;     EAX - number of bytes read
;
Readfile proc near
        Pushad
        Mov esi,edi
        Xor ebp,ebp                ;notre compteur.
        Add esi,codeaddx           ;transforme en adresse linaire.
        Mov edi,Offset Regs
        Lea ebx,[ecx+esi]          ;ebx=ecx+esi
        Mov edx,ecx                ;edx = nbre d'octets  lire.
        Cmp ebx,100000h            ;si le buffer est en mmoire tendue...
        Ja short Readlong          ;on lit dans un buffer temporaire.
        Mov eax,esi
        Shr eax,4
        And esi,0Fh
        Mov [edi]._ds,ax           ;segment mode reel du buffer
        Mov [edi]._dx,si           ;offset mode reel du buffer.
Readl:
        Mov eax,0FFF0h             ;65520 = max value.
        Cmp eax,edx                ;plus d'un segment  lire ??
        Jbe short Readlf1          ;non, parfait, ce sera vite envoy !
        Mov eax,edx
Readlf1:
        Mov [edi]._cx,ax           ;le nombre d'octets  lire dans regs._cx
        Mov [edi]._ah,3Fh          ;la fonction DOS.
        Mov bl,21h
        Call RealInt
        Movzx ebx,[edi]._ax
        Add ebp,ebx                ;compte les octets lus.
        Sub edx,ebx                ;les enlve aux octets  lire.
        Jbe short Readdone         ;si plus d'octets, alors termin.
        Or ebx,ebx
        Jz short Readdone          ;si on a rien lu, alors fini.
        Add [edi]._ds,0FFFh        ;on passe au segment suivant
        Jmp short Readl            ;et on continue la lecture.

Readlong:
        Xor ecx,ecx
        Sub esi,codeaddx      ;destination = User destination.
        Mov ax,bufferseg
        Mov [edi]._ds,ax      ;segment du buffer.
        Mov [edi]._dx,cx      ;offset du buffer.
Readlongl:
        Mov eax,0FFF0h        ;eax = taille du buffer.
        Cmp eax,edx           ;plus d'un segment  lire ?
        Jbe short Readlonglf1 ;non alors c'est cool !
         Mov eax,edx
Readlonglf1:
        Mov [edi]._cx,ax
        Mov [edi]._ah,3Fh
        Mov bl,21h
        Call RealInt
        Movzx ecx,[edi]._ax
        Add ebp,ecx           ;compte les octets lus en ebp.
        Or ecx,ecx            ;pas d'octets lus ?
        Jz short Readdone     ;alors c'est fini.
        Push edi
        Mov edi,esi
        Sub edx,ecx           ;edx = octets  lire - octets lus.
        Mov esi,bufferbase
        Jna short Readend     ;s'il en reste plus on copie les derniers.
        Shr ecx,2             ;divise par 4
        Rep Movsd             ;copie les octets lus par 4.
        Mov esi,edi
        Pop edi
        Jmp short Readlongl
Readend:
        Mov eax,ecx
        Shr ecx,2
        And al,011b
        Rep Movsd             ;copie les derniers octets lus.
        Mov cl,al
        Rep Movsb
        Pop edi
Readdone:
        Mov [esp]._eax,ebp    ;eax for popad
        Bt dword ptr [edi]._flag,0
        Popad
        Ret
Readfile endp

;
; Write to file
; In:
;   REGS._BX - file handle
;   ESI -> buffer to write from
;   ECX - number of bytes to write
; Out:
;   CF=1 - Error writing file
;     EAX - ?
;   CF=0 - Write went fine
;     EAX - number of bytes written
;
Writefile proc near
        Pushad
        Xor ebp,ebp
        Mov edi,Offset Regs
        Add esi,codeaddx           ;transforme en adresse linaire.
        Lea ebx,[ecx+esi]          ;ebx=ecx+esi
        Mov edx,ecx                ;edx = nbre d'octets  lire.
        Cmp ebx,100000h            ;si le buffer est en mmoire tendue...
        Ja short Writelong         ;on lit dans un buffer temporaire.
        Mov eax,esi
        Shr eax,4
        And esi,0Fh
        Mov [edi]._ds,ax           ;segment mode reel du buffer
        Mov [edi]._dx,si           ;offset mode reel du buffer.
Writel:
        Mov eax,0FFF0h             ;65520
        Cmp eax,edx                ;plus d'un segment  ecrire ??
        Jbe short Writelf1         ;non, parfait, ce sera vite envoy !
        Mov eax,edx
Writelf1:
        Mov [edi]._cx,ax           ;le nombre d'octets  lire dans regs_cx
        Mov [edi]._ah,40h          ;la fonction DOS.
        Mov bl,21h
        Call RealInt
        Movzx ebx,[edi]._ax
        Add ebp,ebx                ;compte les octets crits
        Sub edx,ebx                ;les enlve aux octets  ecrire.
        Jbe short Writedone        ;si plus d'octets, alors termin.
        Or ebx,ebx
        Jz short Writedone         ;si on a rien ecrit, alors fini.
        Add [edi]._ds,0FFFh        ;on passe au segment suivant
        Jmp short Writel           ;et on continue l'ecriture.

Writelong:
        Xor ecx,ecx
        Sub esi,codeaddx      ;destination = User destination.
        Mov ax,bufferseg
        Mov [edi]._ds,ax      ;segment du buffer.
        Mov [edi]._dx,cx      ;offset du buffer.
Writelongl:
        Mov eax,0FFF0h          ;eax = taille du buffer.
        Cmp eax,edx             ;plus d'un segment  lire ?
        Jbe short Writelonglf1  ;non alors c'est cool !
        Mov eax,edx
Writelonglf1:
        Push edi
        Lea ecx,[eax+3]         ;arrondit au suprieur.
        Mov edi,bufferbase
        Shr ecx,2
        Rep Movsd
        Pop edi
        Mov bl,21h
        Mov [edi]._cx,ax
        Mov [edi]._ah,40h
        Call RealInt
        Movzx ecx,[edi]._ax
        Add ebp,ecx           ;compte les octets lus en ebp.
        Sub edx,ecx           ;edx = octets  lire - octets lus.
        Ja short Writelongl   ;s'il en reste plus on copie les derniers.
Writedone:
        Mov [esp]._eax,ebp    ;set eax for popad
        Bt dword ptr [edi]._flag,0
        Popad
        Ret
Writefile Endp
