
;                 MOUSE HANDLER (VESA needed)

; InitMouse     : Initializes the mouse. (see parameters below).
; CloseMouse    : Close Mouse driver and free resources.
; ShowMouse     : Show Mouse Cursor.
; HideMouse     : Hide Mouse Cursor.
; GetMouseInfo  : Return Mouse position and button status.
; ChangeMouse   : Change Mouse Parameters.
;
; LeftFct       : Offset of Fonction to be called when left button pressed.
; RightFct      : Offset of Fonction to be called when right button pressed.
; MoveFct       : Offset of Fonction to be called when mouse moved.


ALIGN 4

Cursor      DD 0           ;offset du sprite curseur.
CursHsze    DD 0
CursVsze    DD 0
CursHmax    DD 0
CursVmax    DD 0

OldHsze     DD 0
OldVsze     DD 0
OldXpos     DD 0
OldYpos     DD 0
BkGrnd      DD 0

LeftFct     DD NulFct      ;Function Called if Left button pressed.
RightFct    DD NulFct      ;                   Right
MoveFct     DD NulFct      ;                   Mouse Moved.

MouseVis    DD -1

MsButtons   DD 0           ;public data.
MsXpos      DD 0           ;
MsYpos      DD 0           ;

MsCallBack  DD 0           ;internal pointers.
MsTerPtr    DP 0

MsReg       IntRegs <?>    ;regs for Mouse Call-Back.
FPUstat     DB 108 DUP (?)

;
; Initializes the Mouse Handler.
; In:
;   ESI - BitMap for Mouse Cursor.
;   EAX - BitMap Width
;   EBX - BitMap Height
;   ECX - X resolution
;   EDX - Y resolution
; Out:
;   CF set on error.
;
InitMouse Proc Near
          Pushad

          Mov Cursor,esi
          Sub ecx,eax
          Sub edx,ebx
          Mov CursHsze,eax
          Mov CursVsze,ebx
          Mov CursHmax,ecx
          Mov CursVmax,edx

          Imul eax,ebx
          Imul eax,byteperpix
          Call AllocMem          ;alloc background memory.
          JC @@NoMouse
          Mov BkGrnd,eax

          Mov byte ptr gs:[00449h],12h     ;simulate 640x480x16 mode
          Xor eax,eax
          Int 33h
          Add ax,1
          Cmc
          JC @@NoMouse

          Push ds cs
          Mov ax,0303h
          Pop ds
          Mov esi,Offset MouseHandler
          Mov edi,Offset MsReg
          Int 31h
          Pop ds
          JC @@NoMouse

          Mov word ptr MsCallBack+2,cx     ;save callback adress
          Mov word ptr MsCallBack,dx

          Mov edi,Offset regs
          Mov [edi]._ax,000Ch
          Mov [edi]._cx,007Fh
          Mov [edi]._es,cx
          Mov [edi]._dx,dx
          Mov bl,33h
          Call RealInt
          Bt [edi]._flag,0
          JC @@NoMouse

          Mov ax,0007h
          Xor ecx,ecx
          Mov edx,[esp]._ecx
          Dec edx
          Int 33h

          Mov ax,0008h
          Xor ecx,ecx
          Mov edx,[esp]._edx
          Dec edx
          Int 33h

          Mov ax,0003h
          Int 33h
          Mov word ptr MsXpos,cx
          Mov word ptr MsYpos,dx

          Cli
          Mov esi,Offset Termine
          Mov edi,Offset MsTerPtr
          Mov eax,Offset MsChainTer
          Xchg eax,[esi]
          STOSD
          Mov ax,cs
          Xchg ax,[esi+4]
          STOSW
          Sti
          Clc
@@NoMouse:
          Popad
 NulFct:  Ret
InitMouse Endp

;
; Uninit the Mouse.
;
CloseMouse Proc Near
           Call HideMouse
           Mov ax,0021h
           Int 33h           ;reset the software.
           Mov ax,0304h
           Mov cx,word ptr MsCallBack+2
           Mov dx,word ptr MsCallBack
           Int 31h           ;free the callback.
           Mov eax,BkGrnd
           Call FreeMem      ;free memory.
           Ret
CloseMouse Endp

;
; Make the mouse visible.
;
ShowMouse Proc Near
          Inc MouseVis
          JNZ short @@NoShow
           Push eax ebx
           Mov eax,MsXpos
           Mov ebx,MsYpos
           Call PrintMouse
           Pop ebx eax
  @@NoShow:
          Ret
ShowMouse Endp

;
; Make the mouse invisible.
;
HideMouse Proc Near
          Push eax
          Mov eax,MouseVis
          Or eax,eax
          JNZ short @@NoHide
            Call RestoreBkGrnd
  @@NoHide:
          Dec eax
          Mov MouseVis,eax
          Pop eax
          Ret
HideMouse Endp


;
; Get Mouse Position and button status.
; Out:
;   EBX - button status (bit0: left button; bit1: right button).
;   ECX - X position.
;   EDX - Y position.
;
GetMouseInfo Proc Near
             Mov ebx,MsButtons
             Mov ecx,MsXpos
             Mov edx,MsYpos
             Ret
GetMouseInfo Endp

;
; Setup Mouse Variables.
; In:
;   ESI - BitMap for Mouse Cursor.
;   EAX - BitMap Width
;   EBX - BitMap Height
;   ECX - X resolution
;   EDX - Y resolution
; Out:
;   CF set on error.
;
ChangeMouse Proc Near
          Push eax ecx edx esi
          Call HideMouse

          Mov eax,BkGrnd
          Call FreeMem           ;free old background memory.

          Mov eax,[esp+12]

          Mov Cursor,esi
          Sub ecx,eax
          Sub edx,ebx
          Mov CursHsze,eax
          Mov CursVsze,ebx
          Mov CursHmax,ecx
          Mov CursVmax,edx

          Imul eax,ebx
          Imul eax,byteperpix
          Call AllocMem          ;alloc background memory.
          JC short @@Fail
          Mov BkGrnd,eax
          Call ShowMouse
          Clc
 @@Fail:  Pop esi edx ecx eax
          Ret
ChangeMouse Endp

; Internal SubRoutines 

MsChainTer  Proc
            Call CloseMouse
            JMP [MsTerPtr]
MsChainTer  Endp

RestoreBkGrnd Proc Near         ;restores the background of the cursor.
            Pushad
            Mov esi,BkGrnd
            Mov eax,OldXpos
            Mov ebx,OldYpos
            Mov ecx,OldHsze
            Mov edx,OldVsze
            Xor ebp,ebp
            Call PutArea
            Popad
            Ret
RestoreBkGrnd Endp

PrintMouse Proc Near
           Pushad
           Mov OldYpos,ebx
           Mov OldXpos,eax
           Mov ecx,CursHsze
           Mov edx,CursVsze
           Sub eax,CursHmax     ;clipping
           JNS @@ClipL
             Xor eax,eax
  @@ClipL: Sub ebx,CursVmax
           JNS @@ClipR
             Xor ebx,ebx
  @@ClipR: Sub ecx,eax
           Sub edx,ebx

           Mov OldHsze,ecx      ;saves the background.
           Mov OldVsze,edx
           Mov eax,OldXpos
           Mov ebx,OldYpos
           Mov edi,BkGrnd
           Call GetArea

           Mov ebp,CursHsze
           Sub ebp,ecx
           Test ProcFlag,2
           Mov esi,Cursor       ;print cursor.
           Jz short @@NoMMX
            Fsave FPUstat
            Call PutSprite
            Frstor FPUstat
           Popad
           Ret
  @@NoMMX:  Call PutSprite
            Popad
            Ret
PrintMouse Endp

MouseHandler Proc Near          ;Internal Mouse Handler.
         Mov eax,[esi]
         Mov es:[edi+2Ah],eax         ;set up return adress cs:ip.
         Add word ptr es:[edi+2Eh],4  ;return stack adjust.
         Mov bx,cs:Data32sel
         Push es edi
         Mov ds,bx
         Mov es,bx
                                ;HERE IT BEGINS
         Movzx eax,MsReg._cx
         Movzx ebx,MsReg._dx
         Cmp eax,MsXpos         ;Mouse Moved ???
         JNZ short @@UpDate
         Cmp ebx,MsYpos
         JZ short @@NoDisp
 @@UpDate:
         Mov MsXpos,eax
         Mov MsYpos,ebx
         Cmp MouseVis,0
         JL short @@NoDisp
           Call RestoreBkGrnd
           Call PrintMouse
           Call [MoveFct]
           JMP short @@NoRight
 @@NoDisp:
         Mov edx,MsReg._ebx
         Mov MsButtons,edx

         Test dl,1
         JZ short @@NoLeft
           Call [LeftFct]
           JMP short @@NoRight
 @@NoLeft:
         Test dl,2
         JZ short @@NoRight
           Call [RightFct]
 @@NoRight:
         Pop edi es
         Iretd
MouseHandler Endp
