ALIGN 4

TempPal         db  768 dup (?)         ; a temporary palette.
BlackPal        db  768 dup (0)         ; a black one.
WhitePal        db  768 dup (63)        ; a white one.
PalBytes        DD 0                    ; Number of bytes to handle.
PalDst          DD 0                    ; source palette pointer.
StartCol        db 0
StepVal         db 0

;
; Set palette
; In:
;   AL  - start color
;   ECX - number of colors to set
;   ESI - palette to load
; Out:
;   NONE
;

SetPalette Proc Near
    push  ecx edx esi

    lea   ecx, [ecx+ecx*2]

    mov   dx, 03C8h
    out   dx, al

    inc   dx
    rep   outsb

    pop   esi edx ecx
    ret
SetPalette Endp

;
; Get palette
; In:
;   AL  - start color
;   ECX - number of colors to get
;   EDI - palette to load
; Out:
;   NONE
;

GetPalette Proc Near
    push  ecx edx edi

    lea   ecx, [ecx+ecx*2]

    mov   dx, 03C7h
    out   dx, al

    add   dl, 02h
    rep   insb

    pop   edi edx ecx
    ret
GetPalette Endp

;
; Fade one palette to another
; In:
;   AL  - start color
;   AH  - speed
;   ECX - number of colors
;   ESI - source palette
;   EDI - destination palette
; Out:
;   NONE (source & destination palette preserved)
;

FadeTo Proc Near
    pushad

    mov   bl, ah                        ; bl = speed

    movzx eax, al
    lea   eax,[eax+eax*2]
    add   esi,eax                       ; start color

    add   edi,eax
    mov   [PalDst],edi

    lea   ecx,[ecx+ecx*2]               ; number of bytes to copy ...
    mov   edi, offset TempPal
    mov   PalBytes,ecx
    rep   movsb                         ; copy source palette to temp palette.

    mov   ax, 63
    div   bl
    or    ah, ah                        ; reste ??
    setz  ah
    add   al, ah
    mov   bh, al                        ; bh = compteur de tour.

  @@Bcle:
    mov   ecx, PalBytes
    mov   dx, 03C8h
    mov   al, [esp+001Ch]               ; rcupre la couleur dpart.
    out   dx, al                        ; on commence  la couleur zero.
    mov   esi, offset TempPal
    mov   edi, [PalDst]                 ; rcupre edi.

  @@Another:
    mov   al, [esi]
    mov   ah, [edi]
    mov   dl, ah
    sub   dl, al
    cmp   dl, bl
    jg    @@Plus
      neg   dl
      cmp   dl, bl
      jg    @@Moins
        mov   al,ah

  @@Ecrire:
    mov   [esi], al
    inc   esi
    inc   edi
    dec   ecx
    jnz   @@Another

  @@fini:
    mov   ecx, PalBytes
    mov   esi, offset TempPal
    call  Synchro
    mov   dx, 03C9h
    rep   outsb
    dec   bh
    jnz   @@Bcle

    popad
    ret

  @@Moins:
    sub   al, bl
    jmp   @@Ecrire

  @@Plus:
    add   al,bl
    jmp   @@Ecrire
FadeTo Endp

;
; Initialize Fade palette to another step by step.
; In:
;   AL  - start color
;   AH  - speed
;   ECX - number of colors
;   ESI - source palette
;   EDI - destination palette
; Out:
;   NONE (source & destination palette preserved)
;

InitFadeStep Proc Near
    Push  eax ecx esi edi
    mov   word ptr StartCol, ax         ;StepVal = increment de couleur.

    movzx eax, al
    lea   eax,[eax+eax*2]
    add   esi,eax                       ; start color

    add   edi,eax
    mov   [PalDst],edi

    lea   ecx,[ecx+ecx*2]               ; number of bytes to copy ...
    mov   edi, offset TempPal
    mov   PalBytes,ecx
    rep   movsb                         ; copy source palette to temp palette.

    Pop   edi esi ecx eax
    Ret
InitFadeStep Endp

;
; Fade previous initialized palette one step.
; In:
;   NONE
; Out:
;   NONE
;
FadeStep Proc Near
    pushad

    mov   bl, StepVal                   ; bl = speed

    mov   ecx, PalBytes
    mov   dx, 03C8h
    mov   al, StartCol                  ; rcupre la couleur dpart.
    out   dx, al                        ; on commence  la couleur zero.
    mov   esi, offset TempPal
    mov   edi, [PalDst]                 ; rcupre edi.

  @@Another:
    mov   al, [esi]
    mov   ah, [edi]
    mov   dl, ah
    sub   dl, al
    cmp   dl, bl
    jg    @@Plus
      neg   dl
      cmp   dl, bl
      jg    @@Moins
        mov   al,ah

  @@Ecrire:
    mov   [esi], al
    inc   esi
    inc   edi
    dec   ecx
    jnz   @@Another

  @@fini:
    mov   ecx, PalBytes
    mov   esi, offset TempPal
    mov   dx, 03C9h
    rep   outsb

    popad
    ret

  @@Moins:
    sub   al, bl
    jmp   @@Ecrire

  @@Plus:
    add   al,bl
    jmp   @@Ecrire
FadeStep Endp

;
; Fade screen palette to black (color 0)
; In:
;   AL  - speed
; Out:
;

FadeToBlack Proc Near
    push  eax ebx edi esi ecx

    mov   bl,al

    xor   eax, eax
    mov   edi, offset TempPal
    mov   ecx, 256
    call  GetPalette

    mov   esi, edi
    mov   edi, offset BlackPal
    mov   al, bl
    call  FadeTo

    pop   ecx esi edi ebx eax
    ret
FadeToBlack Endp
