; DESEXT VESA functions 
ALIGN 4
PutScreen   DD offset Putscrbank
Getscreen   DD offset Getscrbank
Putarea     DD offset Putareabank
Getarea     DD offset Getareabank
Putsprite   DD offset Putsprbank

BankList    DD Putscrbank, Getscrbank, Putareabank, Getareabank, Putsprbank
LinList     DD Putscrlin, Getscrlin, Putarealin, Getarealin, Putsprlin
MMXList     DD PutscrMMX, GetscrMMX, Putarealin, Getarealin, PutsprMMX

Bnk16List   DD Putscrbank, Getscrbank, Putarea16B, Getarea16B, Putspr16B
Lin16List   DD Putscrlin, Getscrlin, Putarea16L, Getarea16L, Putspr16L
MMX16List   DD PutscrMMX, GetscrMMX, Putarea16L, Getarea16L, Putspr16M

Bnk32List   DD Putscrbank, Getscrbank, Putarea32B, Getarea32B, Putspr32B
Lin32List   DD Putscrlin, Getscrlin, Putarea32L, Getarea32L, Putspr32L
MMX32List   DD PutscrMMX, GetscrMMX, Putarea32L, Getarea32L, Putspr32M

;
; Set Video Mode, fill ModeInfo if available and init VESA vars.
;  if linear supported, then automatically set linear mode.
;  if MMX unit available, then automatically use MMX.
; In:
;   BL = bit/pixel.
;   CX = X resolution desired.
;   DX = Y resolution desired.
; Out:
;   CF = 0 -> mode set.
;   CF = 1 -> failed.
;   registers preserved, destroys regs._ax, regs._cx, regs._es:di, regs._ss:sp
;
Setvideomode proc near
        Pushad
        Mov ebp,VESAinfo
        Mov esi,modeinfo
        Movzx edi,word ptr [ebp+16]
        Movzx ebp,word ptr [ebp+14]
        Shl edi,4
        Shl edx,16
        Add edi,ebp
        Sub edi,CodeAddx   ;edi = pointer to video modes.
        Mov dx,cx          ;edx = resolution

   @@Next:
        Mov ax,[edi]
        Cmp ax,0FFFFh      ;on s'arrte au dernier mode avec erreur.
        Lea edi,[edi+2]
        JZ @@Abort
        Call Getmodeinfo
        JC short @@Next    ;le mode existe ??
        Cmp edx,[esi+18]   ;bonne rsolution ??
        JNZ short @@Next
        Cmp bl,16
        JNZ @@No16
          Cmp dword ptr [esi+31],05060B05h
          JNZ short @@Next
          Cmp word ptr [esi+35],0005h
          JNZ short @@Next
          JMP short @@Ok
@@No16: Cmp bl,15
        JNZ @@No15
          Cmp dword ptr [esi+31],05050A05h
          JNZ short @@Next
          Cmp word ptr [esi+35],0005h
          JNZ short @@Next
          JMP short @@Ok
@@No15: Cmp bl,[esi+25]    ;nbre de bpp ok ?
        JNZ short @@Next

 @@Ok:  Movzx ebx,bl
        Shr ebx,3
        Mov edi,Offset BankList
        Mov byteperpix,ebx
        Cmp bl,2
        Jnz short @@No16bit
          Add edi,5*3*4
 @@No16bit:
        Cmp bl,4
        Jnz short @@No32bit
          Add edi,5*3*4*2
 @@No32bit:
        Mov ebx,eax
        Movzx eax,word ptr [esi+16]   ;get scan line length.
        Mov scanline,eax
        Mov dword ptr scrresx,edx
        BT dword ptr [esi],7          ;linear available ??
        Mov esi,edi
        JNC short @@SetMode
         Mov eax,linearsvga
         Or eax,eax
         JZ short @@SetMode
          Or bx,4000h
          Test procflag,2
          Lea esi,[esi+4*5]           ;Fast speed !
          JZ short @@SetMode
           Add esi,4*5                ;Ultimate speed !

   @@SetMode:
        Mov edi,Offset PutScreen
        Mov ecx,5
        Rep Movsd                     ;set appropriate functions.

        Mov ax,4F02h
        Int 10h
        Cmp ax,004Fh
        Setnz al
        Add al,0FFh         ;set carry if al <> 0.
        Popad
        Ret

  @@Abort:
        Stc
        Popad
        Ret
Setvideomode endp

;
; Set Scan Line Length (and update ScanLine var).
; In:
;   EAX = New scan line length in PIXELS !!.
; Out:
;   CF = 0 -> scan line changed.
;             EAX = actual scan line in pixels.
;   CF = 1 -> failed.
;   registers preserved.
;
Setscanline proc near
        Push eax ebx ecx edx
        Mov ecx,eax
        Xor ebx,ebx
        Mov eax,4F06h
        Int 10h
        Movzx ecx,cx
        Mov word ptr scanline,bx         ;never more than 0FFFFh
        Mov [esp+12],ecx                 ;value for pop eax.
        Add ah,0FFh                      ;cf set if ah <> 00h
        Pop edx ecx ebx eax
        Ret
Setscanline endp

;
; Fill ModeInfo with informations for a VESA video mode.
; In:
;   AX = Video mode number.
; Out:
;   CF = 0 -> ModeInfo filled.
;   CF = 1 -> failed.
;   registers preserved, destroys regs._ax, regs._cx, regs._es:di, regs._ss:sp
;
Getmodeinfo proc near
        Push eax ebx ecx edi
        Mov edi,Offset regs
        Mov [edi]._cx,ax
        Mov [edi]._ax,4F01h
        Mov eax,modeinfo
        Add eax,codeaddx
        Mov ebx,eax
        Shr eax,4
        And bx,000Fh
        Mov [edi]._es,ax
        Mov [edi]._di,bx
        Mov bl,10h
        Call RealInt
        Cmp [edi]._ax,004Fh
        Setnz al
        Add al,0FFh
        Pop edi ecx ebx eax
        Ret
Getmodeinfo endp

;
; Copy Screen into linear frame buffer
; In:
;   ESI - offset of virtual screen.
;   ECX - number of Dword to Copy.
;   EDI - offset in linear frame buffer. (dword aligned for compatibility).
; Out:
;   registers preserved, MMX registers destroyed.
;
Putscrbank proc near
         Push eax ebx ecx edx esi edi   ;saves only used regs.
         Mov edx,edi                    ; start address in edx
         Mov eax,ecx                    ; save number of dwords in eax
         And edi,0000FFFFh              ; get start offset in window
         Shr edx,16                     ; get first win number

         Mov ecx,edi                    ; calc number of dword in first win.
         Mov ebx,gfxofs                 ; GFXOfs in ebx
         Not cx

         Inc ecx
         Shr ecx,2
@@Nextwin:
         Call vbepaging                 ; set window
         Inc edx                        ; next 64k in video memory

         Cmp ecx,eax                    ; check if count is less than a win.
         Jbe short @@Sizeok
         Mov ecx,eax
@@Sizeok:
         Add edi,ebx                    ; set linear address (+ GFXOfs)
         Sub eax,ecx                    ; update dwords counter
         Rep Movsd                      ; fast copy of memory block

         Xor edi,edi                    ; edi -> 0
         Or eax,eax                     ; if counter <> 0
         Mov ecx,16384
         Jnz short @@Nextwin            ; then next window.

         Pop edi esi edx ecx ebx eax
         Ret
Putscrbank endp

PutscrMMX proc near
          Push edi esi ecx
          Add edi,linearsvga
          Test cl,01111b       ;64 byte multiple ?
          Jnz short No64align
MMXcopy:
          Shr ecx,4
 @@Nxt64:
          DB 0Fh,06Fh,06h           ;          MovQ MM0,[esi]
          DB 0Fh,06Fh,04Eh,8        ;          MovQ MM1,[esi+8]
          DB 0Fh,06Fh,056h,16       ;          MovQ MM2,[esi+16]
          DB 0Fh,06Fh,05Eh,24       ;          MovQ MM3,[esi+24]
          Add edi,64
          DB 0Fh,06Fh,066h,32       ;          MovQ MM4,[esi+32]
          DB 0Fh,06Fh,06Eh,40       ;          MovQ MM5,[esi+40]
          DB 0Fh,06Fh,076h,48       ;          MovQ MM6,[esi+48]
          DB 0Fh,06Fh,07Eh,56       ;          MovQ MM7,[esi+56]
          DB 0Fh,07Fh,047h,-64      ;          MovQ [edi-64],MM0
          DB 0Fh,07Fh,04Fh,-56      ;          MovQ [edi-56],MM1
          DB 0Fh,07Fh,057h,-48      ;          MovQ [edi-48],MM2
          DB 0Fh,07Fh,05Fh,-40      ;          MovQ [edi-40],MM3
          Lea esi,[esi+64]
          DB 0Fh,07Fh,067h,-32      ;          MovQ [edi-32],MM4
          DB 0Fh,07Fh,06Fh,-24      ;          MovQ [edi-24],MM5
          Dec ecx
          DB 0Fh,07Fh,077h,-16      ;          MovQ [edi-16],MM6
          DB 0Fh,07Fh,07Fh,-8       ;          MovQ [edi-8],MM7
          Jnz short @@Nxt64
          Pop ecx esi edi
          Ret
PutscrMMX endp

Putscrlin proc near
         Push edi esi ecx
         Add edi,linearsvga
No64align:
         Rep Movsd
         Pop ecx esi edi
         Ret
Putscrlin endp

;
; Copy Linear Video Memory into virtual screen.
; In:
;   ESI - offset in video memory. (dword aligned for compatibility).
;   ECX - number of Dword to Copy.
;   EDI - offset of virtual screen.
; Out:
;   registers preserved.
;
GetscrMMX proc near
          Push edi esi ecx
          Add esi,linearsvga
          Test cl,01111b       ;64 byte multiple ?
          Jz short MMXcopy
          Rep Movsd
          Pop ecx esi edi
          Ret
GetscrMMX endp

Getscrlin proc near
         Push edi esi ecx
         Add esi,linearsvga
         Rep Movsd
         Pop ecx esi edi
         Ret
Getscrlin endp

Getscrbank proc near
         Push eax ebx ecx edx esi edi   ;saves only used regs.
         Mov edx,esi                    ; start address in edx
         Mov eax,ecx                    ; save number of dwords in eax
         And esi,0000FFFFh              ; get start offset in window
         Shr edx,16                     ; get first win number

         Mov ecx,esi                    ; calc number of dword in first win.
         Mov ebx,gfxofs                 ; GFXOfs in ebx
         Not cx
         Inc ecx
         Shr ecx,2
@@Nextwin:
         Call vbepaging                 ; set window
         Inc edx                        ; next 64k in video memory

         Cmp ecx,eax                    ; check if count is less than a win.
         Jbe short @@Sizeok
         Mov ecx,eax
@@Sizeok:
         Add esi,ebx                    ; set linear address (+ GFXOfs)
         Sub eax,ecx                    ; update dwords counter
         Rep Movsd                      ; fast copy memory block

         Xor esi,esi                    ; esi -> 0
         Or eax,eax                     ; if counter <> 0
         Mov ecx,16384
         Jnz short @@Nextwin            ; then next window.

         Pop edi esi edx ecx ebx eax
         Ret
Getscrbank endp

;
; Copy area from virtual screen into linear frame buffer
; In:
;   ESI - source data
;   EAX - X dest
;   EBX - Y dest
;   ECX - area width
;   EDX - area height
;   EDI - sprite increment.
; Out:
;   registers preserved.
;
Putarea16B proc near           ;Slow (compatibility only)
         Pushad
         Add ecx,ecx
         Add eax,eax
         Add edi,edi
         JMP short PutBnk
Putarea16B endp

Putarea32B proc near           ;Slow (compatibility only)
         Pushad
         Shl ecx,2
         Shl eax,2
         Shl edi,2
         JMP short PutBnk
Putarea32B endp

Putareabank proc near
         Pushad
PutBnk:
         Push edi
         Mov ebp,scanline
         Imul ebx,ebp
         Add ebx,eax           ;ebx=offset in vidmem.
         Sub ebp,ecx           ;ebp=line increment.
         Xchg ebx,edx          ;edx=offset in vidmem; ebx=area height.
         Mov eax,ecx           ;eax=area width
@@Nextwin:
         Rol edx,16
         Call vbepaging        ;set page
         Inc edx               ;next page in high word of edx.
         Rol edx,16
@@Nextline:
         Movzx edi,dx
         Mov ecx,eax
         Add edi,gfxofs

         Add dx,cx             ;test the overflow.
         Jnc short @@Nooverflow

           Sub cx,dx           ;bytes before ovf...
           Rep Movsb

           Rol edx,16          ;select next window...
           Call vbepaging
           Inc edx
           Rol edx,16

           Mov edi,gfxofs
           Mov cx,dx           ;bytes after ovf.
@@Nooverflow:
         Rep Movsb

         Dec ebx
         Jz short @@Fin

         Add esi,[esp]         ;sprite increment
         Add dx,bp
         Jnc short @@Nextline
         Jmp short @@Nextwin
@@Fin:
         Pop eax
         Popad
         Ret
Putareabank endp

Putarealin proc near
         Pushad

         Mov ebp,scanline
         Add eax,linearsvga           ; get dest address
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax+ebx]
         Mov eax,ecx                  ; save area width in eax
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Shr ecx,2
         Rep Movsd                    ; copy a line of area
         Mov cl,al
         And cl,011b
         Rep Movsb
         Add edi,ebp                  ; go to next line in video memory
         Add esi,ebx
         Dec edx                      ; last line ?
         Mov ecx,eax                  ; get area width
         Jnz short @@Copyline

         Popad
         Ret
Putarealin endp

Putarea16L proc near
         Pushad

         Mov ebp,scanline
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax*2+ebx]
         Mov eax,ecx                  ; save area width in eax
         Add edi,linearsvga           ; get dest address
         Add ecx,ecx
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Mov ecx,eax                  ; get area width
         Shr ecx,1
         Rep Movsd                    ; copy a line of area
         Jnc short @@Skip
          Mov ax,[esi]
          Mov [edi],ax
          Add esi,2
          Add edi,2
 @@Skip: Dec edx                      ; last line ?
         Lea edi,[edi+ebp]            ; go to next line in video memory
         Lea esi,[esi+ebx*2]
         Jnz short @@Copyline

         Popad
         Ret
Putarea16L endp

Putarea32L proc near
         Pushad

         Mov ebp,scanline
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax*4+ebx]
         Mov eax,ecx                  ; save area width in eax
         Add edi,linearsvga           ; get dest address
         Shl ecx,2
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Mov ecx,eax                  ; get area width
         Rep Movsd                    ; copy a line of area
         Dec edx                      ; last line ?
         Lea edi,[edi+ebp]            ; go to next line in video memory
         Lea esi,[esi+ebx*4]
         Jnz short @@Copyline

         Popad
         Ret
Putarea32L endp

;
; Copy area from linear video memory into main memory.
; In:
;   EDI - destination buffer.
;   EAX - X source
;   EBX - Y source
;   ECX - area width
;   EDX - area height
; Out:
;   registers preserved.
;
Getarealin proc near
         Pushad
         Mov ebp,scanline
         Add eax,linearsvga             ; get dest address
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax+ebx]
         Sub ebp,ecx                    ; internal line length
         Mov eax,ecx                    ; save area width in eax
@@Copyline:
         Shr ecx,2
         Rep Movsd                      ; copy a line of area
         Mov cl,al
         And cl,011b
         Rep Movsb
         Add esi,ebp                    ; go to next line in video memory
         Dec edx                        ; last line ?
         Mov ecx,eax                    ; get area width
         Jnz short @@Copyline
         Popad
         Ret
Getarealin endp

Getarea16L proc near
         Pushad
         Mov ebp,scanline
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax*2+ebx]
         Mov eax,ecx                    ; save area width in eax
         Add esi,linearsvga             ; get dest address
         Add ecx,ecx
         Sub ebp,ecx                    ; internal line length
@@Copyline:
         Mov ecx,eax                    ; get area width
         Shr ecx,1
         Rep Movsd                      ; copy a line of area
         Jnc short @@Skip
          Mov ax,[esi]
          Mov [edi],ax
          Add esi,2
          Add edi,2
 @@Skip: Dec edx                        ; last line ?
         Lea esi,[esi+ebp]              ; go to next line in video memory
         Jnz short @@Copyline
         Popad
         Ret
Getarea16L endp

Getarea32L proc near
         Pushad
         Mov ebp,scanline
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax*4+ebx]
         Mov eax,ecx                    ; save area width in eax
         Add esi,linearsvga             ; get dest address
         Shl ecx,2
         Sub ebp,ecx                    ; internal line length
@@Copyline:
         Mov ecx,eax                    ; get area width
         Rep Movsd                      ; copy a line of area
         Dec edx                        ; last line ?
         Lea esi,[esi+ebp]              ; go to next line in video memory
         Jnz short @@Copyline
         Popad
         Ret
Getarea32L endp

Getarea16B proc near           ;Slow (compatibility only)
         Pushad
         Add ecx,ecx
         Add eax,eax
         JMP short GetBnk
Getarea16B endp

Getarea32B proc near           ;Slow (compatibility only)
         Pushad
         Shl ecx,2
         Shl eax,2
         JMP short GetBnk
Getarea32B endp

Getareabank proc near
         Pushad
GetBnk:
         Mov ebp,scanline
         Imul ebx,ebp
         Add ebx,eax           ;ebx=offset in vidmem.
         Sub ebp,ecx           ;ebp=line increment.
         Xchg ebx,edx          ;edx=offset in vidmem; ebx=area height.
         Mov eax,ecx           ;eax=area width
@@Nextwin:
         Rol edx,16
         Call vbepaging        ;set page
         Inc edx               ;next page in high word of edx.
         Rol edx,16
@@Nextline:
         Movzx esi,dx
         Mov ecx,eax
         Add esi,gfxofs

         Add dx,cx             ;test the overflow.
         Jnc short @@Nooverflow

           Sub cx,dx           ;bytes before ovf...
           Rep Movsb

           Rol edx,16          ;select next window...
           Call vbepaging
           Inc edx
           Rol edx,16

           Mov esi,gfxofs
           Mov cx,dx           ;bytes after ovf.
@@Nooverflow:
         Rep Movsb

         Dec ebx
         Jz short @@Fin

         Add dx,bp
         Jnc short @@Nextline
         Jmp short @@Nextwin
@@Fin:
         Popad
         Ret
Getareabank endp

;
; Copy sprite into linear frame buffer
; In:
;   ESI - source data
;   EAX - Xdest
;   EBX - Ydest
;   ECX - sprite width  (multiple de 64bit pour utiliser le MMX)
;   EDX - sprite height
;   EDI - sprite line increment (for sprite clipping) 0=whole sprite.
; Out:
;   registers preserved.
;
PutsprMMX proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,0111b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No8mul
         Add eax,linearsvga
         Imul ebx,ebp
         Lea edi,[eax+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Shr ecx,3
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,74h,0FEh        ;         Pcmpeqb MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
PutsprMMX endp

Putsprlin proc near
         Pushad
         Mov ebp,scanline
No8mul:
         Add eax,linearsvga
         Imul ebx,ebp
         Lea edi,[eax+ebx]
         Sub ebp,ecx
         Mov ebx,[esp]._edi     ; sprite increment.
@@Copyline:
         Mov al,[esi]           ; faster than lodsb if not used with REP
         Or al,al
         Jz short @@Nextpixel
          Mov [edi],al
@@Nextpixel:
         Inc esi
         Inc edi
         Dec ecx
         Jnz short @@Copyline

         Add esi,ebx
         Add edi,ebp
         Dec edx
         Mov ecx,[esp]._ecx
         Jnz short @@Copyline
         Popad
         Ret
Putsprlin endp

Putspr16M proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,011b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No4mul
         Imul ebx,ebp
         Add ecx,ecx
         Lea edi,[eax*2+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Add edi,linearsvga
         Shr ecx,3
         Add eax,eax
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,75h,0FEh        ;         Pcmpeqw MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
Putspr16M endp

Putspr16L proc near
         Pushad
         Mov ebp,scanline
No4mul:
         Imul ebx,ebp
         Add ecx,ecx
         Sub ebp,ecx
         Lea edi,[eax*2+ebx]
         Mov ebx,[esp]._edi       ; sprite increment.
         Add edi,linearsvga
         Xor eax,eax
@@Copyline:
         Mov ecx,[esp]._ecx
@@Copybyte:
         Mov ax,[esi]
         Or eax,eax
         Jz short @@Nextpixel
          Mov [edi],ax
@@Nextpixel:
         Dec ecx
         Lea esi,[esi+2]
         Lea edi,[edi+2]
         Jnz short @@Copybyte
         Dec edx
         Lea esi,[esi+ebx*2]
         Lea edi,[edi+ebp]
         Jnz short @@Copyline
         Popad
         Ret
Putspr16L endp

Putspr32M proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,01b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No2mul
         Imul ebx,ebp
         Shl ecx,2
         Lea edi,[eax*4+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Shr ecx,3
         Add edi,linearsvga
         Shl eax,2
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,76h,0FEh        ;         Pcmpeqd MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
Putspr32M endp

Putspr32L proc near
         Pushad
         Mov ebp,scanline
No2mul:
         Imul ebx,ebp
         Shl ecx,2
         Sub ebp,ecx
         Lea edi,[eax*4+ebx]
         Mov ebx,[esp]._edi
         Add edi,linearsvga
@@Copyline:
         Mov ecx,[esp]._ecx
@@Copybyte:
         Mov eax,[esi]
         Or eax,eax
         Jz short @@Nextpixel
          Mov [edi],eax
@@Nextpixel:
         Dec ecx
         Lea esi,[esi+4]
         Lea edi,[edi+4]
         Jnz short @@Copybyte
         Dec edx
         Lea esi,[esi+ebx*4]
         Lea edi,[edi+ebp]
         Jnz short @@Copyline
         Popad
         Ret
Putspr32L endp

Putsprbank proc near
         Pushad                ;[esp]._ecx = sprite width.
         Mov ebp,scanline
         Imul ebx,ebp
         Add ebx,eax           ;ebx=offset in vidmem.
         Sub ebp,ecx           ;ebp=line increment.
         Xchg ebx,edx          ;edx=offset in vidmem; ebx=area height.
@@Nextwin:
         Rol edx,16
         Call vbepaging        ;set page
         Inc edx               ;next page in high word of edx.
         Rol edx,16
@@Nextline:
         Movzx edi,dx
         Mov ecx,[esp]._ecx    ;ecx = sprite width. (from stack)
         Add edi,gfxofs

         Add dx,cx             ;test the overflow.
         Jnc short @@Nooverflow

           Sub cx,dx           ;bytes before ovf...
  @@Nxtbyte1:
           Mov al,[esi]
           Or al,al
           Jz short @@Skip1
           Mov [edi],al
  @@Skip1: Inc esi
           Inc edi
           Dec ecx
           Jnz short @@Nxtbyte1

           Rol edx,16          ;select next window...
           Call vbepaging
           Inc edx
           Rol edx,16

           Mov edi,gfxofs      ;new page: restart at begining of memory.
           Mov cx,dx           ;bytes after ovf.
@@Nooverflow:
  @@Nxtbyte2:
         Mov al,[esi]
         Or al,al
         Jz short @@Skip2
         Mov [edi],al
  @@Skip2:
         Inc esi
         Inc edi
         Dec ecx
         Jnz short @@Nxtbyte2

         Dec ebx
         Jz short @@Fin

         Add esi,[esp]._edi    ;sprite increment
         Add dx,bp
         Jnc short @@Nextline
         Jmp short @@Nextwin
@@Fin:
         Popad
         Ret
Putsprbank endp

Putspr16B proc near
         Pushad
         Add edi,edi
         Add ecx,ecx
         Push edi ecx          ;[esp] = sprite width. [esp+4] = sprite inc.
         Mov ebp,scanline
         Imul ebx,ebp
         Lea ebx,[ebx+eax*2]   ;ebx=offset in vidmem.
         Sub ebp,ecx           ;ebp=line increment.
         Xchg ebx,edx          ;edx=offset in vidmem; ebx=area height.
@@Nextwin:
         Rol edx,16
         Call vbepaging        ;set page
         Inc edx               ;next page in high word of edx.
         Rol edx,16
@@Nextline:
         Movzx edi,dx
         Mov ecx,[esp]         ;ecx = sprite width. (from stack)
         Add edi,gfxofs

         Add dx,cx             ;test the overflow.
         Jnc short @@Nooverflow

           Sub cx,dx           ;bytes before ovf...
           Shr ecx,1
  @@Nxtbyte1:
           Mov ax,[esi]
           Or ax,ax
           Jz short @@Skip1
           Mov [edi],ax
  @@Skip1: Dec ecx
           Lea esi,[esi+2]
           Lea edi,[edi+2]
           Dec ecx
           Jnz short @@Nxtbyte1

           Rol edx,16          ;select next window...
           Call vbepaging
           Inc edx
           Rol edx,16

           Mov edi,gfxofs      ;new page: restart at begining of memory.
           Mov cx,dx           ;bytes after ovf.
@@Nooverflow:
         Shr ecx,1
  @@Nxtbyte2:
         Mov ax,[esi]
         Or ax,ax
         Jz short @@Skip2
         Mov [edi],ax
  @@Skip2:
         Dec ecx
         Lea esi,[esi+2]
         Lea edi,[edi+2]
         Jnz short @@Nxtbyte2

         Dec ebx
         Jz short @@Fin

         Add esi,[esp+4]
         Add dx,bp
         Jnc short @@Nextline
         Jmp short @@Nextwin
@@Fin:
         Pop eax ebx
         Popad
         Ret
Putspr16B endp

Putspr32B proc near
         Pushad
         Shl edi,2
         Shl ecx,2
         Push edi ecx          ;[esp] = sprite width. [esp+4] = sprite inc.
         Mov ebp,scanline
         Imul ebx,ebp
         Lea ebx,[ebx+eax*4]   ;ebx=offset in vidmem.
         Sub ebp,ecx           ;ebp=line increment.
         Xchg ebx,edx          ;edx=offset in vidmem; ebx=area height.
@@Nextwin:
         Rol edx,16
         Call vbepaging        ;set page
         Inc edx               ;next page in high word of edx.
         Rol edx,16
@@Nextline:
         Movzx edi,dx
         Mov ecx,[esp]         ;ecx = sprite width. (from stack)
         Add edi,gfxofs

         Add dx,cx             ;test the overflow.
         Jnc short @@Nooverflow

           Sub cx,dx           ;bytes before ovf...
           Shr ecx,2
  @@Nxtbyte1:
           Mov eax,[esi]
           Or eax,eax
           Jz short @@Skip1
           Mov [edi],eax
  @@Skip1: Dec ecx
           Lea esi,[esi+4]
           Lea edi,[edi+4]
           Jnz short @@Nxtbyte1

           Rol edx,16          ;select next window...
           Call vbepaging
           Inc edx
           Rol edx,16

           Mov edi,gfxofs      ;new page: restart at begining of memory.
           Mov cx,dx           ;bytes after ovf.
@@Nooverflow:
         Shr ecx,2
  @@Nxtbyte2:
         Mov eax,[esi]
         Or eax,eax
         Jz short @@Skip2
         Mov [edi],eax
  @@Skip2:
         Dec ecx
         Lea esi,[esi+4]
         Lea edi,[edi+4]
         Jnz short @@Nxtbyte2

         Dec ebx
         Jz short @@Fin

         Add esi,[esp+4]
         Add dx,bp
         Jnc short @@Nextline
         Jmp short @@Nextwin
@@Fin:
         Pop eax ebx
         Popad
         Ret
Putspr32B endp

;
; Wait Synchro.
;   registers preserved.
;
Synchro proc near
        Push eax edx
        Mov edx,03DAh
 @@WaitEndVBL:
        In al,dx
        And al,8
        Jnz short @@WaitEndVBL
 @@WaitStartVBL:
        In al,dx
        And al,8
        Jz short @@WaitStartVBL
        Pop edx eax
        Ret
Synchro endp
