; DESEXT VESA functions 
ALIGN 4
PutScreen   DD offset putscrlin
Putarea     DD offset Putarealin
Getarea     DD offset Getarealin
Putsprite   DD offset Putsprlin

LinList     DD Putscrlin, Putarealin, Getarealin, Putsprlin
MMXList     DD PutscrMMX, Putarealin, Getarealin, PutsprMMX

Lin16List   DD Putscrlin, Putarea16L, Getarea16L, Putspr16L
MMX16List   DD PutscrMMX, Putarea16L, Getarea16L, Putspr16M

Lin32List   DD Putscrlin, Putarea32L, Getarea32L, Putspr32L
MMX32List   DD PutscrMMX, Putarea32L, Getarea32L, Putspr32M

;
; Set Video Mode, fill ModeInfo if available and init VESA vars.
;  if linear supported, then automatically set linear mode.
;  if MMX unit available, then automatically use MMX.
; In:
;   BL = bit/pixel.
;   CX = X resolution desired.
;   DX = Y resolution desired.
; Out:
;   CF = 0 -> mode set.
;   CF = 1 -> failed. (no linear mode, or no corresponding mode)
;   registers preserved, destroys regs._ax, regs._cx, regs._es:di, regs._ss:sp
;
Setvideomode proc near
        Pushad
        Mov ebp,VESAinfo
        Mov esi,modeinfo
        Movzx edi,word ptr [ebp+16]
        Movzx ebp,word ptr [ebp+14]
        Shl edi,4
        Shl edx,16
        Add edi,ebp
        Sub edi,CodeAddx   ;edi = pointer to video modes.
        Mov dx,cx          ;edx = resolution

   @@Next:
        Mov ax,[edi]
        Cmp ax,0FFFFh      ;on s'arrte au dernier mode avec erreur.
        Lea edi,[edi+2]
        JZ @@Abort
        Call Getmodeinfo
        JC short @@Next    ;le mode existe ??
        Cmp edx,[esi+18]   ;bonne rsolution ??
        JNZ short @@Next
        Cmp bl,16
        JNZ @@No16
          Cmp dword ptr [esi+31],05060B05h
          JNZ short @@Next
          Cmp word ptr [esi+35],0005h
          JNZ short @@Next
          JMP short @@Ok
@@No16: Cmp bl,15
        JNZ @@No15
          Cmp dword ptr [esi+31],05050A05h
          JNZ short @@Next
          Cmp word ptr [esi+35],0005h
          JNZ short @@Next
          JMP short @@Ok
@@No15: Cmp bl,[esi+25]    ;nbre de bpp ok ?
        JNZ short @@Next

 @@Ok:  Movzx ebx,bl
        Shr ebx,3
        Mov edi,Offset LinList
        Mov byteperpix,ebx
        Cmp bl,2
        Jnz short @@No16bit
          Add edi,4*2*4
 @@No16bit:
        Cmp bl,4
        Jnz short @@No32bit
          Add edi,4*2*4*2
 @@No32bit:
        Mov ebx,eax
        Movzx eax,word ptr [esi+16]   ;get scan line length.
        Mov scanline,eax
        Mov dword ptr scrresx,edx
        BT dword ptr [esi],7          ;linear available ??
        Mov esi,edi
        JNC short @@Abort
         Mov eax,linearsvga
         Or eax,eax
         JZ short @@Abort
          Or bx,4000h
          Test procflag,2
          JZ short @@SetMode
           Add esi,4*4                ;Ultimate speed ! (MMX)

   @@SetMode:
        Mov edi,Offset Putarea
        Mov ecx,4
        Rep Movsd                     ;set appropriate functions.

        Mov ax,4F02h
        Int 10h
        Cmp ax,004Fh
        Setnz al
        Add al,0FFh         ;set carry if al <> 0.
        Popad
        Ret

  @@Abort:
        Stc
        Popad
        Ret
Setvideomode endp

;
; Set Scan Line Length (and update ScanLine var).
; In:
;   EAX = New scan line length in PIXELS !!.
; Out:
;   CF = 0 -> scan line changed.
;             EAX = actual scan line in pixels.
;   CF = 1 -> failed.
;   registers preserved.
;
Setscanline proc near
        Push eax ebx ecx edx
        Mov ecx,eax
        Xor ebx,ebx
        Mov eax,4F06h
        Int 10h
        Movzx ecx,cx
        Mov word ptr scanline,bx         ;never more than 0FFFFh
        Mov [esp+12],ecx                 ;value for pop eax.
        Add ah,0FFh                      ;cf set if ah <> 00h
        Pop edx ecx ebx eax
        Ret
Setscanline endp

;
; Fill ModeInfo with informations for a VESA video mode.
; In:
;   AX = Video mode number.
; Out:
;   CF = 0 -> ModeInfo filled.
;   CF = 1 -> failed.
;   registers preserved, destroys regs._ax, regs._cx, regs._es:di, regs._ss:sp
;
Getmodeinfo proc near
        Push eax ebx ecx edi
        Mov edi,Offset regs
        Mov [edi]._cx,ax
        Mov [edi]._ax,4F01h
        Mov eax,modeinfo
        Add eax,codeaddx
        Mov ebx,eax
        Shr eax,4
        And bx,000Fh
        Mov [edi]._es,ax
        Mov [edi]._di,bx
        Mov bl,10h
        Call RealInt
        Cmp [edi]._ax,004Fh
        Setnz al
        Add al,0FFh
        Pop edi ecx ebx eax
        Ret
Getmodeinfo endp

;
; Copy Screen into linear frame buffer
; In:
;   ESI - offset of virtual screen.
;   ECX - number of Dword to Copy.
;   EDI - offset in linear frame buffer. (dword aligned for performance).
; Out:
;   registers preserved, MMX registers destroyed.
;
PutscrMMX proc near
          Push edi esi ecx
          Add edi,linearsvga
          Sub ecx,16
          Js short @@Less64
 @@Nxt64:
           DB 0Fh,06Fh,06h           ; MovQ MM0,[esi]
           DB 0Fh,06Fh,04Eh,8        ; MovQ MM1,[esi+8]
           DB 0Fh,06Fh,056h,16       ; MovQ MM2,[esi+16]
           DB 0Fh,06Fh,05Eh,24       ; MovQ MM3,[esi+24]
           Add edi,64
           DB 0Fh,06Fh,066h,32       ; MovQ MM4,[esi+32]
           DB 0Fh,06Fh,06Eh,40       ; MovQ MM5,[esi+40]
           DB 0Fh,06Fh,076h,48       ; MovQ MM6,[esi+48]
           DB 0Fh,06Fh,07Eh,56       ; MovQ MM7,[esi+56]
           Sub ecx,16
           DB 0Fh,07Fh,047h,-64      ; MovQ [edi-64],MM0
           DB 0Fh,07Fh,04Fh,-56      ; MovQ [edi-56],MM1
           DB 0Fh,07Fh,057h,-48      ; MovQ [edi-48],MM2
           DB 0Fh,07Fh,05Fh,-40      ; MovQ [edi-40],MM3
           Lea esi,[esi+64]
           DB 0Fh,07Fh,067h,-32      ; MovQ [edi-32],MM4
           DB 0Fh,07Fh,06Fh,-24      ; MovQ [edi-24],MM5
           DB 0Fh,07Fh,077h,-16      ; MovQ [edi-16],MM6
           DB 0Fh,07Fh,07Fh,-8       ; MovQ [edi-8],MM7
          Jns short @@Nxt64
          Add ecx,16
          DB 0Fh,77h                 ; EEMS
          Jnz short @@Remain
          Ret
 @@Less64:
          Add ecx,16
 @@Remain:
          Rep Movsd
          Pop ecx esi edi
          Ret
PutscrMMX endp

Putscrlin proc near
         Push edi esi ecx
         Add edi,linearsvga
         Rep Movsd
         Pop ecx esi edi
         Ret
Putscrlin endp

;
; Copy area from virtual screen into linear frame buffer
; In:
;   ESI - source data
;   EAX - X dest
;   EBX - Y dest
;   ECX - area width
;   EDX - area height
;   EDI - sprite increment
; Out:
;   registers preserved.
;
Putarealin proc near
         Pushad

         Mov ebp,scanline
         Add eax,linearsvga           ; get dest address
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax+ebx]
         Mov eax,ecx                  ; save area width in eax
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Shr ecx,2
         Rep Movsd                    ; copy a line of area
         Mov cl,al
         And cl,011b
         Rep Movsb
         Add edi,ebp                  ; go to next line in video memory
         Add esi,ebx
         Dec edx                      ; last line ?
         Mov ecx,eax                  ; get area width
         Jnz short @@Copyline

         Popad
         Ret
Putarealin endp

Putarea16L proc near
         Pushad

         Mov ebp,scanline
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax*2+ebx]
         Mov eax,ecx                  ; save area width in eax
         Add edi,linearsvga           ; get dest address
         Add ecx,ecx
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Mov ecx,eax                  ; get area width
         Shr ecx,1
         Rep Movsd                    ; copy a line of area
         Jnc short @@Skip
          Mov ax,[esi]
          Mov [edi],ax
          Add esi,2
          Add edi,2
 @@Skip: Dec edx                      ; last line ?
         Lea edi,[edi+ebp]            ; go to next line in video memory
         Lea esi,[esi+ebx*2]
         Jnz short @@Copyline

         Popad
         Ret
Putarea16L endp

Putarea32L proc near
         Pushad

         Mov ebp,scanline
         Imul ebx,ebp                 ; in edi
         Lea edi,[eax*4+ebx]
         Mov eax,ecx                  ; save area width in eax
         Add edi,linearsvga           ; get dest address
         Shl ecx,2
         Mov ebx,[esp]._edi
         Sub ebp,ecx                  ; internal line length

@@Copyline:
         Mov ecx,eax                  ; get area width
         Rep Movsd                    ; copy a line of area
         Dec edx                      ; last line ?
         Lea edi,[edi+ebp]            ; go to next line in video memory
         Lea esi,[esi+ebx*4]
         Jnz short @@Copyline

         Popad
         Ret
Putarea32L endp

;
; Copy area from linear video memory into main memory.
; In:
;   EDI - destination buffer.
;   EAX - X source
;   EBX - Y source
;   ECX - area width
;   EDX - area height
; Out:
;   registers preserved.
;
Getarealin proc near
         Pushad
         Mov ebp,scanline
         Add eax,linearsvga             ; get dest address
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax+ebx]
         Sub ebp,ecx                    ; internal line length
         Mov eax,ecx                    ; save area width in eax
@@Copyline:
         Shr ecx,2
         Rep Movsd                      ; copy a line of area
         Mov cl,al
         And cl,011b
         Rep Movsb
         Add esi,ebp                    ; go to next line in video memory
         Dec edx                        ; last line ?
         Mov ecx,eax                    ; get area width
         Jnz short @@Copyline
         Popad
         Ret
Getarealin endp

Getarea16L proc near
         Pushad
         Mov ebp,scanline
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax*2+ebx]
         Mov eax,ecx                    ; save area width in eax
         Add esi,linearsvga             ; get dest address
         Add ecx,ecx
         Sub ebp,ecx                    ; internal line length
@@Copyline:
         Mov ecx,eax                    ; get area width
         Shr ecx,1
         Rep Movsd                      ; copy a line of area
         Jnc short @@Skip
          Mov ax,[esi]
          Mov [edi],ax
          Add esi,2
          Add edi,2
 @@Skip: Dec edx                        ; last line ?
         Lea esi,[esi+ebp]              ; go to next line in video memory
         Jnz short @@Copyline
         Popad
         Ret
Getarea16L endp

Getarea32L proc near
         Pushad
         Mov ebp,scanline
         Imul ebx,ebp                   ; in edi
         Lea esi,[eax*4+ebx]
         Mov eax,ecx                    ; save area width in eax
         Add esi,linearsvga             ; get dest address
         Shl ecx,2
         Sub ebp,ecx                    ; internal line length
@@Copyline:
         Mov ecx,eax                    ; get area width
         Rep Movsd                      ; copy a line of area
         Dec edx                        ; last line ?
         Lea esi,[esi+ebp]              ; go to next line in video memory
         Jnz short @@Copyline
         Popad
         Ret
Getarea32L endp

;
; Copy sprite into linear frame buffer
; In:
;   ESI - source data
;   EAX - Xdest
;   EBX - Ydest
;   ECX - sprite width  (multiple de 8 pour utiliser le MMX)
;   EDX - sprite height
;   EBP - sprite increment
; Out:
;   registers preserved.
;
PutsprMMX proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,0111b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No8mul
         Add eax,linearsvga
         Imul ebx,ebp
         Lea edi,[eax+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Shr ecx,3
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,74h,0FEh        ;         Pcmpeqb MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
PutsprMMX endp

Putsprlin proc near
         Pushad
         Mov ebp,scanline
No8mul:
         Add eax,linearsvga
         Imul ebx,ebp
         Lea edi,[eax+ebx]
         Sub ebp,ecx
         Mov ebx,[esp]._edi     ; sprite increment.
@@Copyline:
         Mov al,[esi]           ; faster than lodsb if not used with REP
         Or al,al
         Jz short @@Nextpixel
          Mov [edi],al
@@Nextpixel:
         Inc esi
         Inc edi
         Dec ecx
         Jnz short @@Copyline

         Add esi,ebx
         Add edi,ebp
         Dec edx
         Mov ecx,[esp]._ecx
         Jnz short @@Copyline
         Popad
         Ret
Putsprlin endp

Putspr16M proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,011b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No4mul
         Imul ebx,ebp
         Add ecx,ecx
         Lea edi,[eax*2+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Add edi,linearsvga
         Shr ecx,3
         Add eax,eax
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,75h,0FEh        ;         Pcmpeqw MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
Putspr16M endp

Putspr16L proc near
         Pushad
         Mov ebp,scanline
No4mul:
         Imul ebx,ebp
         Add ecx,ecx
         Sub ebp,ecx
         Lea edi,[eax*2+ebx]
         Mov ebx,[esp]._edi       ; sprite increment.
         Add edi,linearsvga
         Xor eax,eax
@@Copyline:
         Mov ecx,[esp]._ecx
@@Copybyte:
         Mov ax,[esi]
         Or eax,eax
         Jz short @@Nextpixel
          Mov [edi],ax
@@Nextpixel:
         Dec ecx
         Lea esi,[esi+2]
         Lea edi,[edi+2]
         Jnz short @@Copybyte
         Dec edx
         Lea esi,[esi+ebx*2]
         Lea edi,[edi+ebp]
         Jnz short @@Copyline
         Popad
         Ret
Putspr16L endp

Putspr32M proc near        ;MM6 & MM7 destroyed.
         Pushad
         Test cl,01b          ; if not MMX compatible -> slow Sprite.
         Mov ebp,scanline
         Jnz short No2mul
         Imul ebx,ebp
         Shl ecx,2
         Lea edi,[eax*4+ebx]
         Sub ebp,ecx
         Mov eax,[esp]._edi     ; sprite increment.
         Shr ecx,3
         Add edi,linearsvga
         Shl eax,2
         Mov ebx,ecx

@@Copyline:                                             ;6 cycles loop !
         DB 0Fh,0EFh,0FFh       ;         Pxor MM7,MM7      ; MM7 = 0
         DB 0Fh,06Fh,036h       ;         MovQ MM6,[esi]    ; read 8 source bytes.
         Add edi,8
         DB 0Fh,76h,0FEh        ;         Pcmpeqd MM7,MM6   ; create mask
         DB 0Fh,0DBh,07Fh,-8    ;         Pand MM7,[edi-8]  ; apply on dest
         Add esi,8
         DB 0Fh,0EBh,0F7h       ;         Por MM6,MM7       ; mix source & dest
         Dec ecx
         DB 0Fh,07Fh,077h,-8    ;         MovQ [edi-8],MM6  ; write on dest
         Jnz @@Copyline

         Add esi,eax
         Add edi,ebp
         Dec edx
         Mov ecx,ebx
         Jnz short @@Copyline
         Popad
         Ret
Putspr32M endp

Putspr32L proc near
         Pushad
         Mov ebp,scanline
No2mul:
         Imul ebx,ebp
         Shl ecx,2
         Sub ebp,ecx
         Lea edi,[eax*4+ebx]
         Mov ebx,[esp]._edi
         Add edi,linearsvga
@@Copyline:
         Mov ecx,[esp]._ecx
@@Copybyte:
         Mov eax,[esi]
         Or eax,eax
         Jz short @@Nextpixel
          Mov [edi],eax
@@Nextpixel:
         Dec ecx
         Lea esi,[esi+4]
         Lea edi,[edi+4]
         Jnz short @@Copybyte
         Dec edx
         Lea esi,[esi+ebx*4]
         Lea edi,[edi+ebp]
         Jnz short @@Copyline
         Popad
         Ret
Putspr32L endp

;
; Wait Synchro.
;   registers preserved.
;
Synchro proc near
        Push eax edx
        Mov edx,03DAh
 @@WaitEndVBL:
        In al,dx
        And al,8
        Jnz short @@WaitEndVBL
 @@WaitStartVBL:
        In al,dx
        And al,8
        Jz short @@WaitStartVBL
        Pop edx eax
        Ret
Synchro endp
